/*****************************************************************************/
/*																			 */
/* fichier	: univers.c									version : V1.0		 */
/* projet	: univers									date :	22/07/2004	 */
/* par		: aerith														 */
/*																			 */
/* simulation d'un univers 3D d'elements simple								 */
/*																			 */
/*****************************************************************************/

#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<math.h>
#include	<time.h>
#include	<GL/glut.h>

#define		NB_CARLINE		1024
#define		PAUSE			printf("\nPressez \'Entrer\' pour continuer\n");fflush(stdout);getchar()

typedef struct _vecteur {
	float	x;
	float	y;
	float	z;
} vecteur;

typedef struct _quaternion {
	float	x;
	float	y;
	float	z;
	float	a;
} quaternion;

typedef struct _camera {
    vecteur    Pos;
	vecteur    Ang;
} camera;

typedef struct _window {
	int		Name;
	char	Title[256];
	int		SizeX;
	int		SizeY;
} window;

typedef struct _element {
	vecteur    Pos;
	int        Taille;
	int        Densitee;
	int        Charge;
	int        Polaritee;
	vecteur    Move;
} element;

typedef struct _engine {
	float		RotatePas;
	float		TranslatePas;
	int			NbEdgeShere;
	float		Clipping;
} engine;

typedef struct _physique {
	int		    Speed;
	int		    Sensible;
	float		Acceleration;
	int			Pause;
} physique;

void	ElementsLoad(void);
void	ElementsSave(void);
void    ElementsAdd(element Data);
void    ElementsDel(int Num);

void    CalcPhysique(void);

void	LoadParameter(char *FichierInit);
void	InitEngine(int argc, char *argv[]);
GLvoid	Reshape(GLsizei width, GLsizei height);
void	Draw(void);
void	AfficheText(int x, int y, char *Text);

void	GestionClavier(unsigned char key, int x, int y);
void	GestionSpecial(int key, int x, int y);

char 	*gettok(char * Data, int Pos, char *Car);

unsigned int	NbElem;
element			*Elements;
camera			Camera;
window			WinEngine;
engine			OptEngine;
physique        Physique;

int		main(int argc, char *argv[], char *envp[])
{
	ElementsLoad();

	LoadParameter("engine.init");

	InitEngine(argc, argv);

	return 0;
}

/****************************************************************************/
/* System physique														 	*/
/****************************************************************************/
void	CalcPhysique(void)
{
	Camera.Ang.x	+= OptEngine.RotatePas;
	Camera.Ang.y	+= OptEngine.RotatePas/2;
	Camera.Ang.z	+= OptEngine.RotatePas/4;



}

/****************************************************************************/
/* Outils																 	*/
/****************************************************************************/
char *	gettok(char *Data, int Pos, char *Car)
{
	static char	Data2[NB_CARLINE];
	char *	Token;
	int		i;
	strcpy(Data2, Data);
	for(Token = strtok(Data2, Car), i = 0; Token && (i < Pos); ++i, Token = strtok(0, Car));
	return Token;
}

/*****************************************************************************/
/* gestion db d'elements													 */
/*****************************************************************************/
void	ElementsLoad(void)
{
	FILE *	Fichier;
	char	Data[NB_CARLINE];
	
    free(Elements);
	NbElem = 0;
	*Data = 0;

	printf("\nOpen Data-Base...\n");

	Fichier = fopen("Elements.udb", "r");
	if(!Fichier)
	{
		printf("\nError Elements.udb not exist\n");
		PAUSE;
		exit(2);
	}
	while(!feof(Fichier))
	{
		if(fgets(Data, NB_CARLINE, Fichier))
            NbElem++;
	}
	fclose(Fichier);

	printf("%i Elements found, Loading...\n", NbElem);

    Elements = calloc(NbElem + 1, sizeof(element));
    NbElem = 0;
	Fichier = fopen("Elements.udb", "r");
	if(!Fichier)
	{
		printf("\nError Elements.udb not exist\n");
		PAUSE;
		exit(2);
	}
	while(!feof(Fichier))
	{
		if(fgets(Data, NB_CARLINE, Fichier))
		{
			Elements[NbElem].Pos.x       = atof(gettok(Data, 0, "\t ,"));
			Elements[NbElem].Pos.y       = atof(gettok(Data, 1, "\t ,"));
			Elements[NbElem].Pos.z       = atof(gettok(Data, 2, "\t ,"));
			Elements[NbElem].Taille      = atoi(gettok(Data, 3, "\t ,"));
			Elements[NbElem].Densitee    = atoi(gettok(Data, 4, "\t ,"));
			Elements[NbElem].Charge      = atoi(gettok(Data, 5, "\t ,"));
			Elements[NbElem].Polaritee   = atoi(gettok(Data, 6, "\t ,"));
			Elements[NbElem].Move.x      = atof(gettok(Data, 7, "\t ,"));
			Elements[NbElem].Move.y      = atof(gettok(Data, 8, "\t ,"));
			Elements[NbElem].Move.z      = atof(gettok(Data, 9, "\t ,"));
			NbElem++;
		}
	}
	fclose(Fichier);

	printf("Successfull Loaded\n");
}

void	ElementsSave(void)
{
	char			Data[NB_CARLINE];
    unsigned int    i;
	FILE			*Fichier;

	printf("\nWrite to Data-Base...\n");

	Fichier = fopen("Elements.udb", "w+");
	if(!Fichier)
	{
		printf("\nError Elements.udb not exist\n");
		PAUSE;
		exit (2);
	}
	for(i = 0; i < NbElem; i++)
	{
    	sprintf(Data, "%.2f,%.2f,%.2f \t %i %i %i %i \t %.2f,%.2f,%.2f\n",
            Elements[i].Pos.x,
            Elements[i].Pos.y,
            Elements[i].Pos.z,
            Elements[i].Taille,
            Elements[i].Densitee,
            Elements[i].Charge,
            Elements[i].Polaritee,
            Elements[i].Move.x,
            Elements[i].Move.y,
            Elements[i].Move.z
        );
		fwrite(Data, sizeof(char), strlen(Data), Fichier);
	}
	fclose(Fichier);

	printf("Save Successfull\n");
    Physique.Pause = 1;
}

void	ElementsAdd(element Data)
{
    Elements[NbElem].Pos.x      = Data.Pos.x;
    Elements[NbElem].Pos.y      = Data.Pos.y;
    Elements[NbElem].Pos.z      = Data.Pos.z;
    Elements[NbElem].Taille     = Data.Taille;
    Elements[NbElem].Densitee   = Data.Densitee;
    Elements[NbElem].Charge     = Data.Charge;
    Elements[NbElem].Polaritee  = Data.Polaritee;
    Elements[NbElem].Move.x     = Data.Move.x;
    Elements[NbElem].Move.y     = Data.Move.y;
    Elements[NbElem].Move.z     = Data.Move.z;
	NbElem++;
	ElementsSave();
	ElementsLoad();
}

void	ElementsDel(int Num)
{
    unsigned long i;
	for(i = Num; i < NbElem; i++);
    {
        Elements[i].Pos.x      = Elements[i+1].Pos.x;
        Elements[i].Pos.y      = Elements[i+1].Pos.y;
        Elements[i].Pos.z      = Elements[i+1].Pos.z;
        Elements[i].Taille     = Elements[i+1].Taille;
        Elements[i].Densitee   = Elements[i+1].Densitee;
        Elements[i].Charge     = Elements[i+1].Charge;
        Elements[i].Polaritee  = Elements[i+1].Polaritee;
        Elements[i].Move.x     = Elements[i+1].Move.x;
        Elements[i].Move.y     = Elements[i+1].Move.y;
        Elements[i].Move.z     = Elements[i+1].Move.z;
    }
	NbElem--;
	ElementsSave();
	ElementsLoad();
}

/*****************************************************************************/
/* moteur de rendu 3D														 */
/*****************************************************************************/
void	LoadParameter(char *FichierInit)
{
	FILE *	Fichier;
	char	Data[NB_CARLINE];
	clock_t Start, Finish;

	printf("\nLoading Parameters...\n");
	Start = clock();

	//parametre par defaut
	Camera.Pos.x	= 0;
	Camera.Pos.y	= 0;
	Camera.Pos.z	= 10;
	Camera.Ang.x	= 0;
	Camera.Ang.y	= 0;
	Camera.Ang.z   	= 0;
	OptEngine.RotatePas		= 1;
	OptEngine.TranslatePas	= 1;
	OptEngine.NbEdgeShere	= 8;
	OptEngine.Clipping		= 200;
	Physique.Speed			= 100;
	Physique.Sensible		= 1;
	Physique.Acceleration	= 10;
	Physique.Pause			= 0;
	WinEngine.SizeX	= 320;
	WinEngine.SizeY	= 240;
	strcpy(WinEngine.Title, "univer");

	//chargement des parametre par fichier
	Data[0] = 0;
	Fichier = fopen(FichierInit, "r");
	if(!Fichier)
	{
		printf("erreur ouverture fichier %s\n", FichierInit);
		PAUSE;
		exit(2);
	}
	while(!feof(Fichier))
	{
		if(fgets(Data, NB_CARLINE, Fichier) != NULL)
		{
			if(!(Data[0] == '/' && Data[1] == '/') && Data[0] != '\n')
			{
				if(strcmp(gettok(Data, 0, "=: \t\n\""), "Option") == 0)
				{
					OptEngine.RotatePas		= atof(gettok(Data, 1, " \t\n"));
					OptEngine.TranslatePas	= atof(gettok(Data, 2, " \t\n"));
					OptEngine.NbEdgeShere	= atoi(gettok(Data, 3, " \t\n"));
					OptEngine.Clipping		= atof(gettok(Data, 4, " \t\n"));
				}
				if(strcmp(gettok(Data, 0, "=: \t\n\""), "Physique") == 0)
				{
					Physique.Speed		    = atoi(gettok(Data, 1, " \t\n"));
					Physique.Sensible	    = atoi(gettok(Data, 2, " \t\n"));
					Physique.Acceleration	= atof(gettok(Data, 3, " \t\n"));
					Physique.Pause			= atoi(gettok(Data, 4, " \t\n"));
				}
				if(strcmp(gettok(Data, 0, "=: \t\n\""), "CamPos") == 0)
				{
					Camera.Pos.x	= atof(gettok(Data, 1, " \t\n"));
					Camera.Pos.y	= atof(gettok(Data, 2, " \t\n"));
					Camera.Pos.z	= atof(gettok(Data, 3, " \t\n"));
				}
				if(strcmp(gettok(Data, 0, "=: \t\n\""), "CamAng") == 0)
				{
					Camera.Ang.x	= atof(gettok(Data, 1, " \t\n"));
					Camera.Ang.y	= atof(gettok(Data, 2, " \t\n"));
					Camera.Ang.z	= atof(gettok(Data, 3, " \t\n"));
				}
				if(strcmp(gettok(Data, 0, "=: \t\n\""), "Window") == 0)
				{
					WinEngine.SizeX	= atoi(gettok(Data, 1, " \t\n"));
					WinEngine.SizeY	= atoi(gettok(Data, 2, " \t\n"));
				}
				if(strcmp(gettok(Data, 0, "\t\n\""), "Title") == 0)
					strcpy(WinEngine.Title, gettok(Data, 1, "\t\n\""));
			}
		}
	}
	fclose(Fichier);
	
	if(WinEngine.Name)
	{
		glutSetWindowTitle(WinEngine.Title);
		glutReshapeWindow(WinEngine.SizeX, WinEngine.SizeY);
		glutPositionWindow(50, 50);
	}

	printf("Window : %i*%i\n", WinEngine.SizeX, WinEngine.SizeY);
	Finish = clock();
	printf("Initialized in %f ms\n", (Finish - Start) / CLOCKS_PER_SEC);
    Physique.Pause = 1;
}

void	InitEngine(int argc, char *argv[])
{
	//init fenetre
	glutInit(&argc,argv);
	glutInitWindowSize(WinEngine.SizeX, WinEngine.SizeY);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	WinEngine.Name = glutCreateWindow(WinEngine.Title);

	//init opengl
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glMatrixMode(GL_MODELVIEW);

	//init fonctions
	glutReshapeFunc(Reshape);
	glutDisplayFunc(Draw);
	glutSpecialFunc(GestionSpecial);
	glutKeyboardFunc(GestionClavier);

	glutMainLoop();	
}

GLvoid Reshape(GLsizei w, GLsizei h)
{
	if (h==0)
		h=1;
	glViewport(0,0,w,h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f,(GLfloat)w/(GLfloat)h,1.0f,OptEngine.Clipping);
	glMatrixMode(GL_MODELVIEW);
}

void	Draw(void)
{
	unsigned int	i;
	clock_t			Start = clock(), Finish;
	static int		FpsCt = 0;
	char			Text[512];
	int		LightPos[4]	= {Camera.Pos.x, Camera.Pos.y, Camera.Pos.z, 1};
	int		MatSpec[4]  = {1,1,1,1};
	
	glMatrixMode(GL_MODELVIEW);
	
	glMaterialiv(GL_FRONT_AND_BACK,GL_SPECULAR,MatSpec);  
	glMateriali(GL_FRONT_AND_BACK,GL_SHININESS,100);  
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	
	if(Camera.Ang.x > 360)  Camera.Ang.x -= 360;
	if(Camera.Ang.x < -360) Camera.Ang.x += 360;
	if(Camera.Ang.y > 360)  Camera.Ang.y -= 360;
	if(Camera.Ang.y < -360) Camera.Ang.y += 360;
	if(Camera.Ang.z > 360)  Camera.Ang.z -= 360;
	if(Camera.Ang.z < -360) Camera.Ang.z += 360;
	
	glPushMatrix();
	glTranslated(-1, 1, -3);
	glutSolidCube(0.1);
	glPopMatrix();

	
	glPushMatrix();
    glTranslated(-Camera.Pos.x, -Camera.Pos.y, -Camera.Pos.z);
    glRotated(Camera.Ang.x,1,0,0);
    glRotated(Camera.Ang.y,0,1,0);
   	glRotated(Camera.Ang.z,0,0,1);
	glLightiv(GL_LIGHT0,GL_POSITION,LightPos);
	glPopMatrix();


	for(i = 0; i < NbElem; i++)
	{
		glPushMatrix();
	    glTranslated(Camera.Pos.x, Camera.Pos.y, Camera.Pos.z);
	    glRotated(Camera.Ang.x,1,0,0);
	    glRotated(Camera.Ang.y,0,1,0);
    	glRotated(Camera.Ang.z,0,0,1);
		glTranslated(Elements[i].Pos.x, Elements[i].Pos.y, Elements[i].Pos.z);
		glutSolidSphere(Elements[i].Taille, OptEngine.NbEdgeShere, OptEngine.NbEdgeShere);
		glPopMatrix();
	}
        
	glutSwapBuffers();
	glutPostRedisplay();

	Finish = clock();
	FpsCt += Finish - Start;
	if(FpsCt > Physique.Speed)
	{
		FpsCt = 0;
		if(!Physique.Pause)
			CalcPhysique();
		sprintf(Text, "%s [%i elements][%i fps]", WinEngine.Title, NbElem, (CLOCKS_PER_SEC / (Finish - Start + 1)));
		glutSetWindowTitle(Text);
		system("CLS");
		printf("[pos : %.2f, %.2f, %.2f]\n", Camera.Pos.x, Camera.Pos.y, Camera.Pos.z);
		printf("[Ang : %.2f, %.2f, %.2f]\n", Camera.Ang.x, Camera.Ang.y, Camera.Ang.z);
		printf("[Pause : %s]\n", (Physique.Pause ? "On" : "Off"));
	}
}

void	AfficheText(int x, int y, char *Text)
{
    char *p = Text;

	glColor3f(1, 0, 0 );
	glRasterPos2f(x, y);
    while(*p)
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *p++);
}

void	GestionClavier(unsigned char key, int x, int y)
{
	switch (key)
	{
	case	'2'				:		/* 2 : rx+ */
        Camera.Ang.x	+= OptEngine.RotatePas;
		break;
	case	'8'				:		/* 8 : rx- */
        Camera.Ang.x	-= OptEngine.RotatePas;
		break;
	case	'4'				:		/* 4 : ry+ */
        Camera.Ang.y	+= OptEngine.RotatePas;
		break;
	case	'6'				:		/* 6 : ry- */
        Camera.Ang.y	-= OptEngine.RotatePas;
		break;
	case	'1'				:		/* 1 : rz+ */
        Camera.Ang.z	+= OptEngine.RotatePas;
		break;
	case	'3'				:		/* 3 : rz- */
        Camera.Ang.z	-= OptEngine.RotatePas;
		break;
	case	's'				:		/* s : save database */
        ElementsSave();
		break;
	case	'l'				:		/* l : load database */
        ElementsLoad();
		break;
	case	'r'				:		/* r : reload parameter */
        LoadParameter("engine.init");
		break;
	case	'f'				:		/* f : fullscreen */
        glutFullScreen();
		break;
	case	'w'				:		/* w : window */
		glutReshapeWindow(WinEngine.SizeX, WinEngine.SizeY);
		glutPositionWindow(50, 50);
		break;
	case	'p'				:		/* p : pause */
		if(Physique.Pause)
			Physique.Pause = 0;
		else
			Physique.Pause = 1;
		break;
	}
}

void GestionSpecial(int key, int x, int y)
{
	switch (key)
	{
	case	GLUT_KEY_LEFT       :		/* droite : tx- */
		Camera.Pos.x -= OptEngine.TranslatePas;
		break;
	case	GLUT_KEY_RIGHT	    :		/* gauche : tx+ */
		Camera.Pos.x += OptEngine.TranslatePas;
		break;
	case	GLUT_KEY_PAGE_DOWN	:		/* bas : ty- */
		Camera.Pos.y -= OptEngine.TranslatePas;
		break;
	case	GLUT_KEY_PAGE_UP	:    	/* haut : ty+ */
		Camera.Pos.y += OptEngine.TranslatePas;
		break;
	case	GLUT_KEY_DOWN	    :	    /* page bas : ty- */
		Camera.Pos.z -= OptEngine.TranslatePas;
		break;
	case	GLUT_KEY_UP         :		/* page haut : ty+ */
		Camera.Pos.z += OptEngine.TranslatePas;
		break;
	}
}

